#!/usr/bin/env perl

use strict;
use warnings;

use Readonly;
use IO::File;
use List::Util qw/max/;

use Bio::EnsEMBL::MiscFeature;
use Bio::EnsEMBL::MiscSet;
use Bio::EnsEMBL::Attribute;
use Maize::ScriptHelper;
use Maize::Clonepath;

Readonly my $CONVERSION_FACTOR => 4900;

Readonly my $BAC_MAP_MISC_SET_CODE  => 'acc_bac_map';
Readonly my $HICF_MAP_MISC_SET_CODE => 'hicf_bac_map';

Readonly my %MISC_SET_PARAMETERS => (
    $BAC_MAP_MISC_SET_CODE => {
        name   => 'Accessioned BAC map',
        desc   => 'List of mapped and accessioned BAC clones',
        maxlen => 250000,
    },
    $HICF_MAP_MISC_SET_CODE => {
        name   => 'HICF BAC map',
        desc   => 'Accessioned BAC clones on the HICF FPC map',
        maxlen => 250000,
    },
);

Readonly my $CLONE_NAME_ATTRIB        => 'name';
Readonly my $CLONE_CONTIG_ATTRIB      => 'superctg';
Readonly my $CLONE_INNER_START_ATTRIB => 'inner_start';
Readonly my $CLONE_INNER_END_ATTRIB   => 'inner_end';
Readonly my $CLONE_FPSIZE_ATTRIB      => 'fp_size';
Readonly my $CLONE_STATE_ATTRIB       => 'state';
Readonly my $CLONE_ACCESSION_ATTRIB   => 'embl_acc';
Readonly my $CLONE_LENGTH_ATTRIB      => 'seq_len';
Readonly my $CLONE_STATUS_ATTRIB      => 'seqstatus';
Readonly my $CLONE_EXTERNAL_ATTRIB    => 'attrib';
Readonly my $CLONE_ANNOTATION_ATTRIB  => 'annotation';

Readonly my $ACCESSIONED_STATE => '12:Accessioned';
Readonly my $EXTERNAL_STATUS   => 'false';
Readonly my $ANNOTATION_STATUS => 'II';
Readonly my $SEQUENCE_STATUS   => 'FULLTOP';

# This is used as a place holder for values that will be populated
# using standard clonepath scripts
Readonly my $PLACEHOLDER => 'Placeholder';

my $helper = undef;

MAIN: {
    $helper = Maize::ScriptHelper->new();
    $helper->load_configuration(
        'mtp'     => { 'format'   => 'mtp=s', 'required' => 1 },
        'species' => { 'required' => 1 }
    );

    my $hicf_clones = get_hicf_clones();
    load_clones_into_ensembl($hicf_clones);
}

=head2 get_hicf_clones

    Returns a list of HICF clones

=cut

sub get_hicf_clones {
    my $mtp_filename = $helper->{mtp};
    my $file         = IO::File->new($mtp_filename, 'r');
    my @clones       = ();
    my $clone_count  = 0;
    while (my $line = <$file>) {
        chomp $line;
        my ($clone_name, $chromosome, $agarose_contig,
            $left_band,  $right_band, $note,
            $accession,  $secondary,  $length
        ) = split(',', $line, 9);
        next unless ($note eq 'HICF');
        next unless ($left_band  =~ /^\d+$/);
        next unless ($right_band =~ /^\d+$/);
        $length ||= 0;
        my ($ctg_start, $ctg_end)
            = map { $_ * $CONVERSION_FACTOR } ($left_band, $right_band);
        my $fpc   = "ctg$agarose_contig";
        my $clone = Clone->new(
            'clone_name'   => $clone_name,
            'contig'       => $fpc,
            'chromosome'   => $chromosome,
            'contig_start' => $ctg_start,
            'contig_end'   => $ctg_end,
            'accession'    => $accession,
            'length'       => $length,
        );
        push @clones, $clone;
        $clone_count++;
    }

    $helper->info("Got $clone_count HICF clones\n");
    $helper->info("Longest reported length: ",
        List::Util::max(map { $_->length } @clones));
    return \@clones;
}

=head2 load_clones_into_ensembl

    Loads the HICF clones into Ensembl

=cut

sub load_clones_into_ensembl {
    my ($clones) = @_;

    my $ensembl = $helper->get_ensembl_adaptor();

    my $misc_feature_adaptor = $ensembl->get_adaptor('MiscFeature');
    my $slice_adaptor        = $ensembl->get_adaptor('Slice');
    my $misc_set_adaptor     = $ensembl->get_adaptor('MiscSet');
    my $bac_map_misc_set
        = fetch_misc_set($misc_set_adaptor, $BAC_MAP_MISC_SET_CODE);
    my $hicf_map_misc_set
        = fetch_misc_set($misc_set_adaptor, $HICF_MAP_MISC_SET_CODE);
    for my $clone (@$clones) {
        my ($existing_misc_feature) = @{
            $misc_feature_adaptor->fetch_all_by_attribute_type_value(
                $CLONE_NAME_ATTRIB, $clone->clone_name)
            };
        if (defined $existing_misc_feature) {
            $helper->info(
                "Clone @{[ $existing_misc_feature->display_id ]} already exists. Skipping"
            );
            next;
        }
        $helper->info(
            "Clone @{[ $clone->clone_name ]} not in database. Creating new feature."
        );
        my $fpc_slice
            = $slice_adaptor->fetch_by_region('fpc', $clone->contig);
        $helper->debug("Clone lies on FPC contig ",
            slice_to_string($fpc_slice));
        my $chrom_slice = project_to_chromosome($fpc_slice);
        $helper->debug("Contig is on chromosome ",
            slice_to_string($chrom_slice));

        my $clone_start  = $clone->contig_start + $chrom_slice->start;
        my $clone_end    = $clone->contig_end + $chrom_slice->start;
        $helper->debug("Clone mapped to chromosome at [$clone_start $clone_end]");
        my $misc_feature = Bio::EnsEMBL::MiscFeature->new(
            -START  => $clone->contig_start,
            -END    => $clone->contig_end,
            -STRAND => 1,
            -SLICE  => $chrom_slice
        );
        $misc_feature->add_MiscSet($bac_map_misc_set);
        $misc_feature->add_MiscSet($hicf_map_misc_set);
        add_clone_attributes($misc_feature, $clone);
        $misc_feature_adaptor->store($misc_feature);
    }
}

=head2 fetch_misc_set

    Fetches a misc_set from Ensembl or creates one if it doesn't exist

=cut

sub fetch_misc_set {
    my ($misc_set_adaptor, $misc_set_code) = @_;

    my $misc_set = $misc_set_adaptor->fetch_by_code($misc_set_code);

    if (!defined $misc_set) {
        my $params = $MISC_SET_PARAMETERS{$misc_set_code};
        $misc_set = Bio::EnsEMBL::MiscSet->new(
            -CODE            => $misc_set_code,
            -NAME            => $params->{name},
            -DESCRIPTION     => $params->{desc},
            -LONGEST_FEATURE => $params->{maxlen},
        );
        $helper->info("Storing new MiscSet [code=$misc_set_code ",
            join(' ', map {"$_='$params->{$_}'"} sort keys %$params), "]");
        $misc_set_adaptor->store($misc_set);
    }
    return $misc_set;
}

=head2 add_clone_attributes

    Add appropriate clone attributes to the misc_feature

=cut

sub add_clone_attributes {
    my ($misc_feature, $clone) = @_;
    my %hash = (
        $CLONE_NAME_ATTRIB        => $clone->clone_name,
        $CLONE_CONTIG_ATTRIB      => $clone->contig,
        $CLONE_INNER_START_ATTRIB => $clone->contig_start,
        $CLONE_INNER_END_ATTRIB   => $clone->contig_end,
        $CLONE_FPSIZE_ATTRIB      => $clone->length,
        $CLONE_STATE_ATTRIB       => $ACCESSIONED_STATE,
        $CLONE_ACCESSION_ATTRIB   => $clone->accession,
        $CLONE_LENGTH_ATTRIB      => $PLACEHOLDER,
        $CLONE_STATUS_ATTRIB      => $PLACEHOLDER,
        $CLONE_EXTERNAL_ATTRIB    => $EXTERNAL_STATUS,
        $CLONE_ANNOTATION_ATTRIB  => $ANNOTATION_STATUS,
    );

    while (my ($code, $value) = each %hash) {
        $misc_feature->add_Attribute(new_Attribute($code, $value));
    }
}

=head2 new_Attribute

    Factory for a given attribute

=cut

sub new_Attribute {
    my ($code, $value) = @_;

    return Bio::EnsEMBL::Attribute->new(-CODE => $code, -VALUE => $value);
}

=head2 project_to_chromosome

    Project a given slice up to the chromosome

=cut

sub project_to_chromosome {
    my $slice = shift;
    my ($chrom_slice)
        = map { $_->to_Slice } @{ $slice->project('chromosome') };
    return $chrom_slice;
}

=head2 slice_to_string

    Returns a string representation of the slice

=cut

sub slice_to_string {
    my $slice = shift;
    return join(q{},
        $slice->seq_region_name, " [", $slice->start, ", ", $slice->end, "]");
}

package Clone;

=head2 new

    Constructor

=cut

sub new {
    my $class  = shift;
    my (%args) = @_;
    my $self   = {};
    bless($self, $class);
    $self->clone_name($args{clone_name});
    $self->chromosome($args{chromosome});
    $self->contig($args{contig});
    $self->contig_start($args{contig_start});
    $self->contig_end($args{contig_end});
    $self->accession($args{accession});
    $self->length($args{length});
    return $self;
}

=head2 clone_name

    Gets and sets the value of clone_name

=cut

sub clone_name {
    my $self = shift;
    if (@_) { $self->{clone_name} = shift }
    return $self->{clone_name};
}

=head2 chromosome

    Gets and sets the value of chromosome

=cut

sub chromosome {
    my $self = shift;
    if (@_) { $self->{CHROMOSOME} = shift }
    return $self->{CHROMOSOME};
}

=head2 accession

    Gets and sets the value of accession

=cut

sub accession {
    my $self = shift;
    if (@_) { $self->{ACCESSION} = shift }
    return $self->{ACCESSION};
}

=head2 contig

    Gets and sets the value of contig

=cut

sub contig {
    my $self = shift;
    if (@_) { $self->{contig} = shift }
    return $self->{contig};
}

=head2 contig_start

    Gets and sets the value of contig_start

=cut

sub contig_start {
    my $self = shift;
    if (@_) { $self->{CONTIG_START} = shift }
    return $self->{CONTIG_START};
}

=head2 contig_end

    Gets and sets the value of contig_end

=cut

sub contig_end {
    my $self = shift;
    if (@_) { $self->{CONTIG_END} = shift }
    return $self->{CONTIG_END};
}

=head2 length

    Gets and sets the value of length

=cut

sub length {
    my $self = shift;
    if (@_) { $self->{LENGTH} = shift }
    return $self->{LENGTH};
}

__END__
