#!/usr/local/bin/perl

=pod

=head1 NAME

classify-gene-models.pl - Classify gene predictions as either TE-related or not

=head1 SYNOPSIS

perl classify-gene-models.pl [options]

Options:
 -? --help
    --man
    --registry_file
    --species

=head1 OPTIONS

Migrates information about sequenced clones into the Ensembl DB.

B<--dryrun>
  Do not make any alterations to the database.

B<-h --help>
  Print a brief help message and exits.

B<-m --man>
  Print man page and exit

B<-r --registry_file>
  Use this Ensembl registry file for database connection info.
  Default is <ENSEMBLHOME>/conf/ensembl.registry

B<-b --bac_species>
  Use this species entry from the registry file for the BAC database [REQUIRED].
  
B<-f --fpc_species>
  Use this species entry from the registry file for the FPC database [REQUIRED].
  
B<-c --conf>
  Reads all configuration information from file. Command-line arguments override.

=head1 DESCRIPTION

B<This program> 

  Loads overgo-clone associations into a core Ensembl DB
  
  Maintained by Shiran Pasternak <shiran@cshl.edu>


=cut

use strict;
use warnings;

use Data::Dumper qw(Dumper);
use File::Basename;
use FindBin qw( $Bin );
use Pod::Usage;
use Getopt::Long;

use Bio::EnsEMBL::DBSQL::DBAdaptor;
use Bio::EnsEMBL::Registry;
use Bio::EnsEMBL::Analysis;
use Bio::EnsEMBL::RepeatConsensus;

use Readonly;

Readonly my $RECAT_DELIMITER => "\t";
Readonly my $CLASS_DELIMITER => "\t";

Readonly my $MOBILE_ELEMENT_CLASS => 'Mobile Element';
Readonly my $RETROELEMENT_CLASS   => 'Class I: Retroelement';

Readonly my @MIPS_CLASS_FIELDS =>
    qw(mips_recat report_class l1_class l2_class l3_class l4_class l5_class l6_class);

MAIN: {
    my ($species, $registry, $recat);
    my ($help, $man);
    my $dbh = undef;
    GetOptions(
        "help|?"     => \$help,
        "man"        => \$man,
        "species=s"  => \$species,
        "registry=s" => \$registry,
        "recat=s"    => \$recat,
    ) or pod2usage(2);
    pod2usage(-verbose => 2) if $man;
    pod2usage(1) if $help;

    # Load the ensembl file
    $species || (print("Need a --species\n") && pod2usage());
    $registry ||= './conf/ensembl.registry';
    for my $file_type (qw/registry recat/) {
        my $file = eval "\$$file_type";
        $file || (print("Need a --$file_type\n") && pod2usage());
        -e $file
            || (print("File $file [$file_type] does not exist\n")
            && pod2usage(1));
        -r $file
            || (print("Cannot read $file [$file_type]\n") && pod2usage(1));
        -f $file
            || (print("File $file [$file_type] is not plain-text\n")
            && pod2usage(1));
        -s $file
            || (print("File $file [$file_type] is empty\n") && pod2usage(1));
    }
    Bio::EnsEMBL::Registry->load_all($registry);

    $dbh = Bio::EnsEMBL::Registry->get_DBAdaptor($species, 'core');
    $dbh
        || (print("No core DB for $species set in $registry\n")
        && pod2usage());

    print("Classifying MIPS repeats for $species\n");

    classify_mips_repeats(
        {   'recat'   => $recat,
            'adaptor' => $dbh,
        }
    );
    exit;
}

my (%adaptors);

sub classify_mips_repeats {
    my ($params) = @_;
    my $recat    = $params->{recat};
    my $adaptor  = $params->{adaptor};

    my ($mips, $mips_classes) = load_mips_data($recat);

    preload_mips_classes(
        { 'adaptor' => $adaptor, 'classes' => $mips_classes });
    my $repeats = get_repeat_consensus_features($params);

    print "Got @{[ scalar @$repeats ]} repeats\n";

    assign_mips_classes(
        {   'adaptor' => $adaptor,
            'mips'    => $mips,
            'repeats' => $repeats,
            'classes' => $mips_classes,
        }
    );
}

sub preload_mips_classes {
    my ($params) = @_;
    my $adaptor  = $params->{'adaptor'};
    my $classes  = $params->{'classes'};

    print "Preloading @{[ scalar keys %$classes ]} as attributes\n";
    my $sql = <<SQL;
replace into attrib_type (attrib_type_id, code, name, description)
 values (null, ?, ?, ?)
SQL
    my $statement = $adaptor->dbc->db_handle->prepare($sql);

    for my $key (sort keys %$classes) {
        my $class = $classes->{$key};
        my $description
            = join($CLASS_DELIMITER, map { $class->{$_} } @MIPS_CLASS_FIELDS);
        my $code = $key;
        print "$key => $description\n";
        next;
        $code =~ s/\.//g;
        $statement->bind_param(1, $code);
        $statement->bind_param(2, "MIPS/REcat Class $key");
        $statement->bind_param(3, $description);
        $statement->execute();
    }
}

sub _adaptor {
    my ($adaptor, $type) = @_;
    return $adaptors{$type} ||= $adaptor->get_adaptor($type);
}

sub load_mips_data {
    my ($recat) = @_;

    my $mips_classes = +{};

    open(my $fh, $recat)
        or die "Cannot open File $recat: !$\n";

    my $fields        = undef;
    my @class_indexes = ();
    my $mips          = {};
    while (my $line = <$fh>) {
        chomp $line;
        if (!defined $fields) {
            $fields = +{};
            my @field_names = split($RECAT_DELIMITER, lc $line);
            for (my $i = 0; $i < scalar @field_names; $i++) {
                $fields->{ $field_names[$i] } = $i;
            }
            @class_indexes = map { $fields->{$_} } @MIPS_CLASS_FIELDS;
            if (!exists($fields->{'seq_id'})) {
                die
                    "[MIPS] Expected field <seq_id> not found in header line of $recat\n";
            }
            next;
        } else {
            my @row = map { $_ =~ s/^\s+//; $_ =~ s/\s+$//; $_ }
                split($RECAT_DELIMITER, $line);
            my $repeat_name = $row[ $fields->{'seq_id'} ];

            my $class
                = +{ map { $_ => $row[ $fields->{$_} ] } @MIPS_CLASS_FIELDS };

            my $key = $row[ $fields->{'mips_recat'} ];
            $mips_classes->{$key}

                # = join($CLASS_DELIMITER, @row[@class_indexes]);
                = $class;

            $mips->{$repeat_name} = $key;
        }
    }

    print "========\n";
    for my $key (sort keys %$mips_classes) {
        print "$key:\n";
        for my $field (@MIPS_CLASS_FIELDS) {
            print "\t$field => [$mips_classes->{$key}->{$field}]\n";
        }
    }

    close $fh;
    return ($mips, $mips_classes);
}

sub get_repeat_consensus_features {
    my ($params) = @_;
    my $adaptor = $params->{adaptor};

    return $adaptor->dbc->db_handle->selectall_arrayref(
        'SELECT * FROM repeat_consensus',
        { Slice => {} });
}

sub assign_mips_classes {
    my ($params) = @_;
    my $adaptor  = $params->{'adaptor'};
    my $repeats  = $params->{'repeats'};
    my $mips     = $params->{'mips'};
    my $classes  = $params->{'classes'};

    my %update_statements = ();

    for my $field (qw/type class/) {
        my $sql = <<SQL;
update repeat_consensus set repeat_${field} = ? where repeat_name = ?
SQL
        $update_statements{$field} = $adaptor->dbc->db_handle->prepare($sql);
    }

    for my $repeat (@$repeats) {
        my $rname = $repeat->{'repeat_name'};
        $rname =~ s/\-int$//;
        my %values = ('class' => undef, 'type' => undef);
        $values{class} = $mips->{$rname};
        if (!defined($values{class})) {
            print "[ERROR] Cannot find class for '$repeat->{repeat_name}'\n";
            next;
        }
        my $mips_class = $classes->{ $values{class} };
        $values{type} = 'MIPS/REcat '
            . (
              $mips_class->{'l1_class'} ne $MOBILE_ELEMENT_CLASS ? 'Other'
            : $mips_class->{'l2_class'} eq $RETROELEMENT_CLASS
            ? 'Class I Retroelements'
            : 'Class II/III Transposable Elements'
            );
        print "Repeat ",
            join(' ',
            map {"$_=$repeat->{\"repeat_$_\"}"} qw(name type class)), " => ",
            join(' ', map {"$_=$values{$_}"} sort keys %values), "\n";
        for my $field (qw/type class/) {
            my $statement = $update_statements{$field};
            $statement->bind_param(1, $values{$field});
            $statement->bind_param(2, $repeat->{'repeat_name'});
            $statement->execute();
        }
    }

}

1;
