#!/usr/local/bin/perl

=head1 NAME

promote-latest-bacs-to-current.pl - Promotes the top version to current

=head1 SYNOPSIS

perl fix-archived-clones.pl [options]

Options:
 -h --help
 -m --man
 -r --registry_file
 -s --species
 -c --conf

=head1 OPTIONS

Resets the archiving scheme in the database. All accessions become versioned.

B<-h --help>
  Print a brief help message and exits.

B<-m --man>
  Print man page and exit

B<-r --registry>
  Use this Ensembl registry file for database connection info.
  Default is <ENSEMBLHOME>/conf/ensembl.registry

B<-s --species>
  Use this species entry from the registry file for the BAC database [REQUIRED].
  
B<-v --verbose>
  Provides verbose output.
  
=head1 DESCRIPTION

B<This program> 

  Resets the archiving scheme in the database. All accessions become versioned.
  
  Maintained by Shiran Pasternak <shiran@cshl.edu>

=cut

use strict;
use warnings;
use Getopt::Long;
use Pod::Usage;

use Bio::EnsEMBL::Registry;

use English;
use FindBin qw($Bin);
use File::Basename qw( dirname );

use Log::Log4perl;

use vars qw($BASEDIR);

use Readonly;
use English qw( -no-match-vars );

my $logger = Log::Log4perl->get_logger('ensembl');

Readonly my $SUSPICIOUS_CLONES_SQL => <<SQL;
select substr(name, 1, 8) as name from seq_region where name like '%.'
SQL

Readonly my $REMOVE_CURRENT_ATTRIBS => <<SQL;
delete seq_region_attrib
  from seq_region_attrib
       left join attrib_type using (attrib_type_id)
 where attrib_type.code = 'current-version'
SQL

Readonly my $REMOVE_SEQ_REGION_SQL => <<SQL;
delete from seq_region where name = ?
SQL

Readonly my $RENAME_SEQ_REGION_SQL => <<SQL;
update seq_region set name = ? where name = ?
SQL

Readonly my $UPDATE_SEQ_REGION_NAME_SQL => <<SQL;
update seq_region set name = ? where seq_region_id = ?
SQL

Readonly my $REMAP_CONTIGS_SQL => <<SQL;
update assembly
   set asm_seq_region_id = (select seq_region_id from seq_region
                             where name = ?)
 where cmp_seq_region_id in (select seq_region_id from seq_region
                              where name like ?)
SQL

BEGIN {

    # Set the perl libraries
    $BASEDIR = dirname($Bin);

    # unshift @INC, $BASEDIR . '/ensembl-live/ensembl/modules';
    Log::Log4perl::init($BASEDIR . '/conf/log4perl.conf');
}

MAIN: {
    my $help = 0;
    my $man  = 0;
    my ($species_name, $registry_file, $verbose);

    my $db_adaptor = undef;

    GetOptions(
        "help|?"     => \$help,
        "man"        => \$man,
        "species=s"  => \$species_name,
        "registry=s" => \$registry_file,
        "verbose"    => \$verbose,
    ) or pod2usage(2);

    pod2usage(-verbose => 2) if $man;
    pod2usage(1) if $help;

    # Validate file paths
    $registry_file ||= $BASEDIR . '/conf/SiteDefs.pm';

    map {
        -e $_ || (warn("File $_ does not exist\n")    && pod2usage(1));
        -r $_ || (warn("Cannot read $_\n")            && pod2usage(1));
        -f $_ || (warn("File $_ is not plain-text\n") && pod2usage(1));
        -s $_ || (warn("File $_ is empty\n")          && pod2usage(1));
    } $registry_file;

    $species_name || (warn("Need a --species\n") && pod2usage(1));
    Bio::EnsEMBL::Registry->load_all($registry_file);

    $db_adaptor = Bio::EnsEMBL::Registry->get_DBAdaptor($species_name, 'core')
        || (warn("No core DB for $species_name set in $registry_file: $@\n")
        && pod2usage(1));

    my $clones = get_all_clones($db_adaptor);

    $logger->info("Got @{[ scalar @$clones ]} clones");
    my %current_accessions = ();

    $logger->info("Fixing clone and contig names");
    remove_all_current_regions($db_adaptor);
    for my $clone (@$clones) {
        my ($accession, $version)
            = $clone->seq_region_name =~ m/^([A-Z]{2}\d{6})(?:\.(\d+))?$/;

        if (!defined $accession) {
            $logger->warn(
                "Something's wrong with Clone @{[$clone->seq_region_name]} (ID=@{[$clone->dbID]})"
            );
            next;
        }
        if (defined $current_accessions{$accession}) {
            my $max_version = $current_accessions{$accession}->{version};
            if ($version > $max_version) {
                $current_accessions{$accession}->{version} = $version;
                $current_accessions{$accession}->{clone}   = $clone;
            }
        } else {
            $current_accessions{$accession} = +{
                'version' => $version,
                'clone'   => $clone
            };
        }
    }

    $logger->info("Setting current clones");
    my $attribute_adaptor = $db_adaptor->get_AttributeAdaptor();
    my $current_attribute = Bio::EnsEMBL::Attribute->new(
        -CODE        => 'current-version',
        -NAME        => 'Current Accession Version',
        -DESCRIPTION => 'Identifies the most recent version of an accession',
        -VALUE       => 'current'
    );
    for my $accession (sort keys %current_accessions) {
        my $clone   = $current_accessions{$accession}->{clone};
        my $version = $current_accessions{$accession}->{version};
        $logger->debug("$accession.$version (clone=$clone)");
        $attribute_adaptor->store_on_Slice($clone, [$current_attribute])
            unless scalar @{ $clone->get_all_Attributes('current-version') };
    }
}

=pod

=head2 get_all_clones
    Returns all clones

=cut

sub get_all_clones {
    my ($adaptor) = @_;

    my $slice_adaptor = $adaptor->get_SliceAdaptor();

    my $clones = $slice_adaptor->fetch_all('clone');
    $logger->debug("Got @{[ scalar @$clones ]} clones");
    return $clones;
}

=head2 remove_all_current_regions

    Removes all current attributes

=cut

sub remove_all_current_regions {
    my ($dbh) = @_;
    update($dbh, $REMOVE_CURRENT_ATTRIBS);
}

=pod

=head2 update
    Updates the DB

=cut

sub update {
    my ($dbh, $sql, @params) = @_;

    my $statement = $dbh->prepare($sql);
    $statement->execute(@params);
}
