
=head1 NAME

Bio::Tools::Run::Search::SubmittedBlast - Class for submitting BLAST offline

=head1 SYNOPSIS

  see Bio::Tools::Run::Search::WuBlast

=head1 DESCRIPTION

An extension of Bio::Tools::Run::Search::WuBlast to dispatch jobs offline
while providing an immediate response to the user. The jobs themselves are
wrapped in $SiteDefs::ENSEMBL_BLASTSCRIPT.

=cut

# Let the code begin...
package Bio::Tools::Run::Search::SubmittedBlast;
use strict;
use File::Copy qw(mv);

use vars qw(@ISA $SPECIES_DEFS);

use Bio::Tools::Run::Search::WuBlast;
use EnsEMBL::Web::SpeciesDefs;
use Sys::Hostname qw(hostname);
use warnings;

@ISA = qw( Bio::Tools::Run::Search::WuBlast );

BEGIN {
    $SPECIES_DEFS = EnsEMBL::Web::SpeciesDefs->new;
}

#----------------------------------------------------------------------

=head2 run

  Arg [1]   : none
  Function  : Dispatches the blast job using the dispatch_sub method
  Returntype: 
  Exceptions: 
  Caller    : 
  Example   : 

=cut

sub run {
    my $self = shift;

    if (    $self->status ne 'PENDING'
        and $self->status ne 'DISPATCHED')
    {
        $self->warn("Wrong status for run: " . $self->status);
    }

    # Apply environment variables, keeping a backup copy
    my %ENV_TMP = %ENV;
    foreach my $env ($self->environment_variable()) {
        my $val = $self->environment_variable($env);
        if (defined $val) { $ENV{$env} = $val }
        else              { delete($ENV{$env}) }
    }

    # Do the deed
    my $command = $self->command;
    $self->dispatch_sub($command);

    $self->debug("BLAST COMMAND: " . $command . "\n");

    # Restore environment
    %ENV = %ENV_TMP;
    return 1;
}

#----------------------------------------------------------------------

=head2 run_blast

  Arg [1]   : None
  Function  : Fires off the blast command (SUPER::run),
              with a pre-repeatmask step
  Returntype: Boolean
  Exceptions:
  Caller    : 
  Example   :

=cut

sub run_blast {
    my $self = shift;
    return $self->SUPER::run();
}

#----------------------------------------------------------------------

=head2 command_sub

  Arg [1]   : None
  Function  : Internal method to generate the shell sub command.
              This command calls the utils/runblast.pm wrapper script 
              rather that the blast command itself
  Returntype: String: $command
  Exceptions:
  Caller    :
  Example   :

=cut

sub command_sub {
    my $self        = shift;
    my $blastscript = $SiteDefs::ENSEMBL_BLASTSCRIPT;
    my $args        = $self->token;
    my $command     = "$blastscript $args";
    return $command;
}

#----------------------------------------------------------------------

=head2 _repeatmask

  Arg [1]   : 
  Function  : 
  Returntype: 
  Exceptions: 
  Caller    : 
  Example   : 

=cut

sub _repeatmask {
    my $self = shift;

    #TODO: expunge SpDefs
    $ENV{BLASTREPEATMASKER} = $SPECIES_DEFS->ENSEMBL_REPEATMASKER;
    return $self->SUPER::_repeatmask(@_);
}

#----------------------------------------------------------------------

=head2 command

  Arg [1]   : None
  Function  : Generate the blast command itself
  Returntype: String: $command
  Exceptions: 
  Caller    : 
  Example   : 

=cut

=pod
sub command {
    my $self = shift;

    if (!-f $self->fastafile) { $self->throw("Need a query sequence!") }

    my $res_file = $self->reportfile;
    if (-f $res_file) {
        $self->warn("A result already exists for $res_file");
        unlink($self->reportfile);
    }

    my $res_file_local = '/tmp/blast_$$.out';

    $ENV{'BLASTMAT'}    || $self->warn("BLASTMAT variable not set");
    $ENV{'BLASTFILTER'} || $self->warn("BLASTFILTER variable not set");
    $ENV{'BLASTDB'}     || $self->warn("BLASTBD variable not set");

    my $database = $self->database
        || $self->throw("No database");

    my $param_str = '';
    foreach my $param ($self->option) {
        my $val = $self->option($param) || '';
        next if $param eq "repeatmask";
        next if $param eq "-RepeatMasker";
        if ($param =~ /=$/) { $param_str .= " $param$val" }
        elsif ($val) { $param_str .= " $param $val" }
        else         { $param_str .= " $param" }
    }
    $param_str =~ s/[;`&|<>\s]+/ /g;
    my $BDP = $SPECIES_DEFS->ENSEMBL_BLAST_DATA_PATH;
    return join(' ',
        $SPECIES_DEFS->ENSEMBL_BLAST_BIN_PATH . "/" . $self->program_path,
        "$BDP/$database", '[[]]', $param_str);
}
=cut

#----------------------------------------------------------------------

=head2 dispatch_sub

  Arg [1]   :
  Function  : Fires off the command
  Returntype:
  Exceptions:
  Caller    : run method
  Example   :

=cut

sub dispatch_sub {
    my $self = shift;
    my $command = shift || die("Need a command to dispatch!");
    my ($ticket) = $self->statefile =~ m#/([^/]+$)#;

    my $blast_dir   = $SPECIES_DEFS->ENSEMBL_TMP_DIR_BLAST;
    my $out_file    = $self->reportfile;
    my $state_file  = $self->statefile;
    my @PARTS       = split /\//, $state_file;
    my $TICKET_NAME = "$PARTS[-3]$PARTS[-2]-$PARTS[-1]";
    my $flag_file   = "$blast_dir/pending/$TICKET_NAME";
    my $sent_file   = "$blast_dir/sent/$TICKET_NAME";
    my $fail_file   = "$state_file.fail";
    my $fasta_file  = $self->fastafile;

    my $repeatmask_command = $SPECIES_DEFS->ENSEMBL_REPEATMASKER;
    my $project_name       = $self->project_name;

    my $command_line = qq(|sh);

    local *SUBMIT;
    if (open(SUBMIT, $command_line)) {
        if (open(FH, ">$sent_file")) {
            print FH "$state_file";
            close FH;
        }
        $self->_init_command_string();
        $self->_add_command(
            qq(touch $out_file),
            qq(touch $fail_file),
            # qq($command >$out_file 2>$fail_file),
            q(status=$?),    ## Store status of BLAST command
            qq(echo '$state_file' > $flag_file),
            q(exit $status)    # Return exit codo of $command!
        );
        print SUBMIT $self->_command_string();
        close SUBMIT;
        if ($? != 0) {
            die("Submit exited with non-zero status - job not submitted\n");
        }
    } else {
        die("Could not exec script: $!\n");
    }
    return 1;
}

=head2 project_name

    The project name

=cut

sub project_name {
    my $self = shift;
    (my $db_name = $self->database)
        =~ s{([A-Z])[a-z]+_([a-z]{3})[a-z]*\..*\.a([-\w+])\.(\w+)\.fa}
            {ENSEMBL.$1$2.$3.$4};
    return join ':', $self->program_name, $db_name,
        $self->seq->length, $self->seq->alphabet;
}

sub _init_command_string {
    my $self = shift;
    $self->{'command_string'} = '';
}

sub _add_command {
    my $self = shift;
    $self->{'command_string'} .= join "\n", @_, '';
}

sub _command_string {
    my $self = shift;
    return $self->{'command_string'};
}

#----------------------------------------------------------------------
sub remove {
    my $self = shift;
    my ($ticket) = $self->statefile =~ m#/([^/]+$)#;
    my $sec = 5;
    local $SIG{ALRM} = sub { die("kill timeout ($sec secs)\n") };

    my $out = "";
    eval {
        alarm($sec);
        my $processes = `ps -C blastn -o pid,command | grep $ticket | cut -c1-6`;
        for my $process (split /\s+/, $processes) {
            $out .= `kill $process`;
        }
        alarm(0);
    };
    if ($@) { die($@) }
    warn("REMOVING $ticket: ", $out);
    $self->SUPER::remove();
}

#----------------------------------------------------------------------

1;
