package Maize::ScriptHelper;

=head1 NAME

Maize::Clonepath

=head1 SYNOPSIS

  use Maize::Clonepath

  
=head1 DESCRIPTION

A utility module for interfacing between the maize clonepath database and
the Ensembl Database

=cut

use strict;
use warnings;

use Getopt::Long;
use Config::General;
use Readonly;
use Log::Log4perl;
use FindBin qw($Bin);
use File::Basename qw( dirname );

use Bio::EnsEMBL::Registry;

BEGIN {
    Log::Log4perl::init("$Bin/../conf/log4perl.conf");
}

Readonly my @REQUIRED_OPTIONS => ('registry');

Readonly our $BAC_DB_CODE => 'bac';
Readonly our $FPC_DB_CODE => 'fpc';

our ($logger);

Readonly my %DEFAULT_OPTIONS =>
    ('registry' => '${basedir}/conf/SiteDefs.pm',);

=pod

=head2 new
    Constructor

=cut

sub new {
    my $class = shift;
    my ($params) = @_;
    $params ||= +{};
    my $self = bless {%$params}, $class;

    $self->{'basedir'} ||= dirname($Bin);

    $logger = Log::Log4perl->get_logger('maize');

    return $self;
}

=pod

=head2 load_configuration
    Loads configuration from configuration file and command-line arguments.

=cut

sub load_configuration {
    my $self         = shift;
    my (%parameters) = @_;
    my $options      = +{};
    my @extra_parameters
        = grep {defined} map { $_->{format} } values %parameters;
    GetOptions($options, "help|h", "man|m", "species|s=s", "registry|r=s",
        "conf|c=s", "printconf", @extra_parameters)
        or die 'Bad options';

    my %defaults = %DEFAULT_OPTIONS;

    my $configuration = new Config::General(
        -DefaultConfig         => \%defaults,
        -ConfigFile            => $options->{'conf'},
        -MergeDuplicateOptions => 1,
        -InterPolateVars       => 0,
    );
    my %config_object = $configuration->getall;
    while (my ($key, $value) = each %config_object) {
        if (!defined($options->{$key})) {
            $options->{$key} = $value;
        }
    }

    for my $default_option (keys %defaults) {
        $options->{$default_option} ||= $defaults{$default_option};
    }

    my @require_errors = ();
    for my $required (@REQUIRED_OPTIONS,
        grep { $parameters{$_}->{required} } keys %parameters)
    {
        if (!defined($options->{$required})) {
            push @require_errors, $required;
        }
    }
    if (scalar @require_errors) {
        for my $param (@require_errors) {
            print STDERR "Option --$param is required\n";
        }
        exit 1;
    }

    # Interpolate nested variables;
    for my $option (keys %$options) {
        my $value = $options->{$option};
        $value =~ s/\$\{(\w+)\}/$self->{$1}/g;
        $options->{$option} = $value;
    }

    if ($options->{'printconf'}) {
        print "CONFIGURATION\n";
        for my $key (grep { $_ ne 'printconf' } sort keys %$options) {
            printf "   \%-20s : \%s\n", $key, $options->{$key};
        }
        exit 0;
    }

    for my $option (keys %$options) {
        $self->{$option} = $options->{$option};
    }
}

=head2 get_ensembl_adaptor

    Initializes and returns an Ensembl DB adaptor

=cut

sub get_ensembl_adaptor {
    my $self = shift;
    Bio::EnsEMBL::Registry->load_all($self->{registry});
    my $adaptor
        = Bio::EnsEMBL::Registry->get_DBAdaptor($self->{species}, 'core');
}

=head2 info

    Logs info

=cut
sub info {
    my $self = shift;
    $logger->info(@_);
}

=head2 warn

    Logs warn

=cut
sub warn {
    my $self = shift;
    $logger->warn(@_);
}

=head2 debug

    Logs debug

=cut
sub debug {
    my $self = shift;
    $logger->debug(@_);
}

1;

=pod

=head1 AUTHOR

Shiran Pasternak E<lt>shiran@cshl.eduE<gt>

=cut

=head1 COPYRIGHT

Copyright (c) 2007 Cold Spring Harbor Laboratory

This library is free software;  you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

