-- Clean out existing coordinate system
delete from coord_system where name = 'fpc_pseudomolecule';
    
-- Change rankings of existing coordinate systems
update coord_system set rank = 3 where name = 'contig';
update coord_system set rank = 2 where name = 'clone';

select @pseudo_cs_id := 3;

-- Create coordinate system for stitched molecule
insert into coord_system (
    coord_system_id, name, version, rank, attrib
) values (
    @pseudo_cs_id, 'fpc_pseudomolecule', 'abutted_BACs', 1, ('toplevel')
);

-- Delete explicit 'toplevel' attribute for individual BACs
delete seq_region_attrib
  from seq_region_attrib
       left join attrib_type using (attrib_type_id)
 where attrib_type.code = 'toplevel';

-- Delete any seq regions on the fpc_pseudomolecule coord system
delete seq_region, assembly, seq_region_attrib
  from seq_region
       left join assembly
              on assembly.asm_seq_region_id = seq_region.seq_region_id
       left join coord_system using (coord_system_id)
       left join seq_region_attrib using (seq_region_id)
 where coord_system.name = 'fpc_pseudomolecule';
 
-- Copy all the chromosome seq_regions from the FPC DB
insert into seq_region
    (seq_region_id, name, coord_system_id, length)
 select null, sr.name, @pseudo_cs_id, sr.length
   from zea_mays_core_46_fpc.seq_region sr
        left join zea_mays_core_46_fpc.coord_system cs using (coord_system_id)
  where cs.name = 'chromosome';

-- Capture the order of BACs on each chromosome into a temporary table
set @order := 0;
create temporary table ordered_bacs as
select acc_att.value as accession,
       chr.name as chromosome,
       bac.length as length,
       @order := @order + 1 as bac_order,
       bac.seq_region_id as bac_region_id,
       chr_sr.seq_region_id as chr_region_id
  from zea_mays_core_46_fpc.misc_feature mf
       left join zea_mays_core_46_fpc.misc_attrib acc_att
              on mf.misc_feature_id = acc_att.misc_feature_id
             and acc_att.attrib_type_id = (
                 select attrib_type_id
                   from zea_mays_core_46_fpc.attrib_type
                  where code = 'embl_acc')
       left join zea_mays_core_46_fpc.seq_region chr
              on chr.seq_region_id = mf.seq_region_id
       left join seq_region bac
              on bac.name = acc_att.value
       left join seq_region chr_sr
              on chr_sr.name = chr.name
 where bac.length is not null
 order by chromosome, mf.seq_region_start;

-- Create assembly associations between the pseudomolecules and the BACs based
-- on order
set @total := 1;
set @this_chr := 0;
insert into assembly (
    asm_seq_region_id, cmp_seq_region_id,
    asm_start, asm_end,
    cmp_start, cmp_end,
    ori
)
select chr_region_id, bac_region_id,
       asm_start, asm_end,
       cmp_start, cmp_end,
       ori
  from (
select chr_region_id,
       bac_region_id,
       (case when @this_chr != chromosome then 1 else @total + 1001 end) as asm_start,
       @total := (case when @this_chr != chromosome then length else @total + length + 1000 end) as asm_end,
       1 as cmp_start,
       length as cmp_end,
       @this_chr := chromosome as this_chr,
       1 as ori
  from ordered_bacs
 order by bac_order) bacs;

-- Update chromosome lengths
---- Capture length based on the maximum end of the assembly
create temporary table chr_lengths as
select seq_region.seq_region_id, max(assembly.asm_end) as length
  from seq_region
       left join assembly on assembly.asm_seq_region_id = seq_region.seq_region_id
       left join coord_system using (coord_system_id)
 where coord_system.name = 'fpc_pseudomolecule'
 group by 1;
 
---- Perform the update
update seq_region, chr_lengths
   set seq_region.length = chr_lengths.length
 where chr_lengths.seq_region_id = seq_region.seq_region_id;

-- Insert toplevel seq region attributes for chromosomes
insert into seq_region_attrib (seq_region_id, attrib_type_id, value)
select seq_region.seq_region_id, attrib_type.attrib_type_id, '1'
  from seq_region
       left join coord_system using (coord_system_id),
       attrib_type
 where attrib_type.code = 'toplevel'
   and coord_system.name = 'fpc_pseudomolecule';
   
-- Create meta information for pseudomolecules
----- Delete previous meta keys
delete from meta
 where meta_key = 'assembly.mapping'
   and meta_value like 'fpc_pseudomolecule%|clone%';

---- Insert new meta rows
insert into meta (meta_id, meta_key, meta_value)
          values (null,
                  'assembly.mapping',
                  'fpc_pseudomolecule:abutted_BACs|clone:BAC_clones');
insert into meta (meta_id, meta_key, meta_value)
          values (null,
                  'assembly.mapping',
                  'fpc_pseudomolecule:abutted_BACs|clone:BAC_clones|contig:BAC_contigs');